import dataclasses
import logging
from pathlib import Path
import time
from typing import List, Optional
from gefyra.cluster.utils import is_operator_running
from gefyra.exceptions import ClusterError


from gefyra.misc.install import synthesize_config_as_dict, synthesize_config_as_yaml
from gefyra.misc.uninstall import (
    remove_all_clients,
    remove_gefyra_crds,
    remove_gefyra_namespace,
    remove_gefyra_rbac,
    remove_remainder_bridges,
)
from gefyra.types import GefyraInstallOptions
from .utils import stopwatch

logger = logging.getLogger("gefyra")

LB_PRESETS = {
    "aws": GefyraInstallOptions(
        service_type="LoadBalancer",
        service_annotations={
            "service.beta.kubernetes.io/aws-load-balancer-type": "nlb",
            "service.beta.kubernetes.io/aws-load-balancer-nlb-target-type": "ip",
            "service.beta.kubernetes.io/aws-load-balancer-scheme": "internet-facing",
        },
    ),
    "gke": GefyraInstallOptions(
        service_type="LoadBalancer",
        service_annotations={},
    ),
}

PRESET_TYPE_MAPPING = {"aws": "remote", "eks": "remote"}


@stopwatch
def install(
    component: Optional[List[str]] = None,
    preset: Optional[str] = None,
    apply: bool = False,
    wait: bool = False,
    kubeconfig: Optional[Path] = None,
    kubecontext: Optional[str] = None,
    **kwargs,
) -> str:
    from gefyra.configuration import ClientConfiguration

    config = ClientConfiguration(kube_config_file=kubeconfig, kube_context=kubecontext)
    if preset:
        presetoptions = LB_PRESETS.get(preset)  # type: ignore
        if not presetoptions:
            raise RuntimeError(
                f"Preset {preset} not available. Available presets are: {', '.join(LB_PRESETS.keys())}"
            )
        _presetoptions = dataclasses.asdict(presetoptions)  # type: ignore
        _presetoptions.update({k: v for k, v in kwargs.items() if v is not None})
        options = GefyraInstallOptions(**_presetoptions)
    else:
        options = GefyraInstallOptions(
            **{k: v for k, v in kwargs.items() if v is not None}
        )
    logger.debug(f"Using options: {options}")
    output = synthesize_config_as_yaml(options=options, components=component)
    if apply:
        import kubernetes

        try:
            ns = config.K8S_CORE_API.read_namespaced_namespace(name=config.NAMESPACE)
            if ns.status.phase.lower() != "active":
                raise ClusterError(
                    f"Cannot apply Gefyra: namespace {config.NAMESPACE} is in {ns.status.phase} state"
                )
        except:  # noqa
            pass

        objects = synthesize_config_as_dict(options=options, components=component)
        for objs in objects:
            logger.debug(objs)
            try:
                if (
                    objs["kind"] == "Deployment"
                    and objs["metadata"]["name"] == "gefyra-operator"
                ):
                    # if this is the operator and the operator is already running, skip the waiting below
                    # as the Gefyra-Ready event will never be emitted
                    wait = not is_operator_running(config)
                kubernetes.utils.create_from_dict(
                    config.K8S_CORE_API.api_client, data=objs
                )
            except kubernetes.utils.FailToCreateError as e:
                logger.debug(e)
                if e.api_exceptions[0].status not in [409, 422]:
                    logger.error(e)
                    raise ClusterError(f"Could not install Gefyra: {e}")
    if apply and wait:
        logger.debug("Waiting for Gefyra to become ready")
        tic = time.perf_counter()
        from kubernetes.watch import Watch

        w = Watch()
        operator_ready = False
        operator_webhook_ready = False
        # block (forever) until Gefyra cluster side is ready
        for event in w.stream(
            config.K8S_CORE_API.list_namespaced_event, namespace=config.NAMESPACE
        ):
            if event["object"].reason in ["Pulling", "Pulled"]:
                logger.info(event["object"].message)
            if event["object"].reason == "Gefyra-Ready":
                operator_ready = True
                logger.debug("Gefyra operator is ready")
            if event["object"].reason == "Gefyra-Webhook-Ready":
                operator_webhook_ready = True
                logger.debug("Gefyra operator webhook is ready")
            if operator_ready and operator_webhook_ready:
                toc = time.perf_counter()
                logger.info(f"Gefyra became ready in {toc - tic:0.4f} seconds")
                break
        # busywait for the operator webhook to become ready
        _i = 0
        while _i < 20:
            webhook_deploy = config.K8S_APP_API.read_namespaced_deployment(
                name="gefyra-operator-webhook", namespace=config.NAMESPACE
            )
            if webhook_deploy.status.ready_replicas == 1:
                break
            else:
                logger.debug("Waiting for the operator webhook to become ready")
                time.sleep(1)
                _i += 1
        else:
            raise ClusterError("Operator webhook did not become ready")
    return output


@stopwatch
def uninstall(
    kubeconfig: Optional[Path] = None, kubecontext: Optional[str] = None, **kwargs
):
    from gefyra.configuration import ClientConfiguration

    config = ClientConfiguration(kube_config_file=kubeconfig, kube_context=kubecontext)
    logger.info("Removing all Gefyra bridges")
    try:
        remove_remainder_bridges(config)
    except Exception as e:
        logger.debug(e)
    logger.info("Removing remainder Gefyra clients")
    try:
        remove_all_clients()
    except Exception as e:
        logger.debug(e)
    logger.info("Removing Gefyra namespace")
    try:
        remove_gefyra_namespace(config)
    except Exception as e:
        logger.debug(e)
    logger.info("Removing Gefyra API extensions")
    try:
        remove_gefyra_crds(config)
    except Exception as e:
        logger.debug(e)
    logger.info("Removing Gefyra RBAC resources")
    try:
        remove_gefyra_rbac(config)
    except Exception as e:
        logger.debug(e)



import click

from gefyra.api import status
from gefyra.cli.utils import standard_error_handler


@click.command("status", help="Get Gefyra's status")
@standard_error_handler
def status_command():
    print(status())



import logging
from time import sleep
from typing import Tuple, TYPE_CHECKING
from gefyra.api.utils import get_workload_type

if TYPE_CHECKING:
    from kubernetes.client.models import V1Pod

from gefyra.configuration import ClientConfiguration


logger = logging.getLogger(__name__)


def get_env_from_pod_container(
    config: ClientConfiguration, pod_name: str, namespace: str, container_name: str
):
    from kubernetes.client import ApiException
    from kubernetes.stream import stream

    retries = 10
    counter = 0
    interval = 1
    while counter < retries:
        try:
            resp = stream(
                config.K8S_CORE_API.connect_get_namespaced_pod_exec,
                pod_name,
                namespace,
                container=container_name,
                command=["env"],
                stderr=True,
                stdin=False,
                stdout=True,
                tty=False,
            )
            return resp
        except ApiException as e:
            # e.status is 0 for some reason
            if "500 Internal Server Error" in e.reason:
                sleep(interval)
                counter += 1
                logger.debug(
                    f"Failed to get env from pod {pod_name} in namespace {namespace} on"
                    f" try {counter}."
                )
            else:
                raise e
    raise RuntimeError(
        f"Failed to get env from pod {pod_name} in namespace {namespace} after"
        f" {retries} tries."
    )


def get_container(pod, container_name: str):
    for container in pod.spec.containers:
        if container.name == container_name:
            return container
    raise RuntimeError(f"Container {container_name} not found.")


def get_container_image(pod, container_name: str):
    container = get_container(pod, container_name=container_name)
    if container.image:
        return container.image
    raise RuntimeError(f"Container {container_name} image could not be determined.")


def get_container_command(pod, container_name: str):
    container = get_container(pod, container_name=container_name)
    res = []
    if container.command:
        res.extend(container.command)
    if container.args:
        res.extend(container.args)
    return " ".join(res)


def get_container_ports(pod, container_name: str):
    container = get_container(pod, container_name=container_name)
    if container.ports:
        return container.ports
    return []


def get_v1pod(
    config: ClientConfiguration,
    pod_name: str,
    namespace: str,
) -> "V1Pod":
    from kubernetes.client import ApiException

    try:
        pod = config.K8S_CORE_API.read_namespaced_pod(pod_name, namespace)
    except ApiException as e:
        if e.status == 404:
            raise RuntimeError(
                f"Pod {pod_name} in namespace {namespace} does not exist."
            )
        else:
            raise e
    return pod


def is_operator_running(config: ClientConfiguration) -> bool:
    from kubernetes.client import ApiException

    try:
        deploy = config.K8S_APP_API.read_namespaced_deployment(
            name="gefyra-operator", namespace=config.NAMESPACE
        )
        return deploy.status.ready_replicas == 1
    except ApiException:
        return False


def retrieve_pod_and_container(
    workload: str, namespace: str, config: ClientConfiguration
) -> Tuple[str, str]:
    from gefyra.cluster.resources import (
        get_pods_and_containers_for_workload,
        get_pods_and_containers_for_pod_name,
    )

    container_name = ""
    workload_type, workload_name = workload.split("/", 1)

    workload_type = get_workload_type(workload_type)

    if "/" in workload_name:
        workload_name, container_name = workload_name.split("/")

    if workload_type != "pod":
        pods = get_pods_and_containers_for_workload(
            config, name=workload_name, namespace=namespace, workload_type=workload_type
        )
    else:
        pods = get_pods_and_containers_for_pod_name(
            config=config, name=workload_name, namespace=namespace
        )

    while len(pods):
        pod_name, containers = pods.popitem()
        if container_name and container_name not in containers:
            raise RuntimeError(
                f"{container_name} was not found for {workload_type}/{workload_name}"
            )
        actual_container_name = container_name or containers[0]
        if pod_ready_and_healthy(config, pod_name, namespace, actual_container_name):
            return pod_name, actual_container_name

    raise RuntimeError(
        f"Could not find a ready pod for {workload_type}/{workload_name}"
    )


def pod_ready_and_healthy(
    config: ClientConfiguration, pod_name: str, namespace: str, container_name: str
):
    pod = config.K8S_CORE_API.read_namespaced_pod_status(pod_name, namespace=namespace)

    container_idx = next(
        i
        for i, container_status in enumerate(pod.status.container_statuses)
        if container_status.name == container_name
    )

    return (
        pod.status.phase == "Running"
        and pod.status.container_statuses[container_idx].ready
        and pod.status.container_statuses[container_idx].started
        and pod.status.container_statuses[container_idx].state.running
        and pod.status.container_statuses[container_idx].state.running.started_at
    )



# https://github.com/yaml/pyyaml/issues/240#issuecomment-1018712495
def str_presenter(dumper, data):
    """configures yaml for dumping multiline strings
    Ref: https://stackoverflow.com/questions/8640959/how-can-i-control-what-scalar-form-pyyaml-uses-for-my-data
    """
    if len(data.splitlines()) > 1:  # check for multiline string
        return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")
    return dumper.represent_scalar("tag:yaml.org,2002:str", data)



class GefyraClientAlreadyExists(RuntimeError):
    pass


class GefyraClientNotFound(RuntimeError):
    pass


class ClientConfigurationError(RuntimeError):
    pass


class MinikubeError(RuntimeError):
    pass


class GefyraBridgeError(RuntimeError):
    pass


class CommandTimeoutError(RuntimeError):
    pass


class GefyraConnectionError(RuntimeError):
    pass


class ClusterError(RuntimeError):
    pass


class PodNotFoundError(RuntimeError):
    pass


class WorkloadNotFoundError(RuntimeError):
    pass



import dataclasses
from time import sleep
import click
from gefyra.cli import console
from gefyra.cli.utils import (
    check_connection_name,
    parse_ip_port_map,
    standard_error_handler,
)
from tabulate import tabulate


@click.command("bridge", help="Establish a Gefyra bridge to a container in the cluster")
@click.option(
    "-N", "--name", help="The name of the container running in Gefyra", required=True
)
@click.option(
    "-p",
    "--ports",
    help="Add forward port mapping in form of <remote_container_port>:<local_container_port>",
    required=True,
    multiple=True,
    callback=parse_ip_port_map,
)
@click.option(
    "-n",
    "--namespace",
    #   help=namespace_help_text,
    default="default",
)
@click.option(
    "-P",
    "--no-probe-handling",
    is_flag=True,
    help="Make Carrier to not handle probes during switch operation",
    default=False,
)
@click.option(
    "--target",
    help=(
        "Intercept the container given in the notation 'resource/name/container'. "
        "Resource can be one of 'deployment', 'statefulset' or 'pod'. "
        "E.g.: --target deployment/hello-nginx/nginx"
    ),
    required=True,
)
@click.option(
    "--connection-name", type=str, default="default", callback=check_connection_name
)
@click.option("--timeout", type=int, default=60, required=False)
@standard_error_handler
def create_bridge(
    name, ports, target, namespace, no_probe_handling, connection_name, timeout
):
    from alive_progress import alive_bar
    from gefyra import api

    print_keys = {
        "name": "NAME",
        "port_mappings": "PORTS",
        "local_container_ip": "LOCAL ADDRESS",
        "target_container": "TARGET CONTAINER",
        "target_pod": "TARGET POD",
        "target_namespace": "NAMESPACE",
    }
    # we are not blocking this call
    _created_bridges = api.bridge(
        name=name,
        ports=ports,
        target=target,
        namespace=namespace,
        handle_probes=no_probe_handling,
        wait=False,
        connection_name=connection_name,
    )
    with alive_bar(
        total=None,
        length=20,
        title=f"Creating the requested bridge(s) (timeout={timeout}))",
        bar="smooth",
        spinner="classic",
        stats=False,
        dual_line=True,
    ) as bar:
        for _i in range(timeout):
            all_bridges = api.list_gefyra_bridges(connection_name=connection_name)[0][1]
            _created_bridges = [
                bridge
                for bridge in all_bridges
                if bridge.name in [_bridge.name for _bridge in _created_bridges]
            ]
            bar.text(
                "\n".join(
                    [f"{bridge.name}: {bridge.state}" for bridge in _created_bridges]
                )
            )
            if all(bridge.state == "ACTIVE" for bridge in _created_bridges):
                break
            else:
                sleep(1)
        bar.text(f"{len(_created_bridges)} bridge(s) active")

    if _created_bridges:
        bridges_print = [
            {
                k: v
                for k, v in dataclasses.asdict(bridge).items()
                if k in print_keys.keys()
            }
            for bridge in _created_bridges
        ]
        console.success(
            "The following bridges have been created. Run 'gefyra list --bridges' to see all bridges."
        )
        click.echo(
            tabulate(
                bridges_print,
                headers=print_keys,
                tablefmt="plain",
            )
        )


@click.command("unbridge", help="Remove a Gefyra bridge (from 'gefyra list --bridges')")
@click.argument("name", required=False)
@click.option(
    "-A",
    "--all",
    help="Unbridge all bridges",
    required=False,
    is_flag=True,
    default=False,
)
@click.option(
    "--connection-name", type=str, default="default", callback=check_connection_name
)
@standard_error_handler
def unbridge(name: str, connection_name: str, all: bool = False):
    from gefyra import api

    if not all and not name:
        console.error("Provide a name or use --all flag to unbridge.")
    if all:
        api.unbridge_all(connection_name=connection_name, wait=True)
    else:
        api.unbridge(connection_name=connection_name, name=name, wait=True)



from dataclasses import fields
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union
import click
from click import ClickException


def standard_error_handler(func):
    def wrapper(*args, **kwargs):
        try:
            result = func(*args, **kwargs)
            return result
        except Exception as e:  # noqa
            ce = ClickException(message=str(e))
            raise ce

    return wrapper


class AliasedCommand(click.Command):
    def __init__(
        self,
        name,
        alias: Optional[Iterable] = None,
        context_settings=None,
        callback=None,
        params=None,
        help=None,
        epilog=None,
        short_help=None,
        options_metavar="[OPTIONS]",
        add_help_option: bool = True,
        no_args_is_help: bool = False,
        hidden: bool = False,
        deprecated: bool = False,
    ) -> None:
        if not alias:
            alias = []
        super().__init__(
            name,
            context_settings,
            callback,
            params,
            help,
            epilog,
            short_help,
            options_metavar,
            add_help_option,
            no_args_is_help,
            hidden,
            deprecated,
        )
        self.alias = alias


class AliasedGroup(click.Group):
    command_class = AliasedCommand

    def get_command(self, ctx, cmd_name):
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        for _, cmd in self.commands.items():
            if hasattr(cmd, "alias") and cmd_name in cmd.alias:
                return cmd
        matches = [x for x in self.list_commands(ctx) if x.startswith(cmd_name)]
        if not matches:
            return None
        elif len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])
        ctx.fail(f"Too many matches: {', '.join(sorted(matches))}")

    def format_commands(self, ctx, formatter) -> None:
        commands = []
        _run_commands = sorted(self.commands.items())
        for subcommand, acmd in _run_commands:
            cmd = self.get_command(ctx, subcommand)
            # What is this, the tool lied about a command.  Ignore it
            if cmd is None:
                continue
            if cmd.hidden:
                continue

            if hasattr(acmd, "alias"):
                alias = ",".join(acmd.alias)
            else:
                alias = None
            commands.append((f"{subcommand} {'(' + alias + ')' if alias else ''}", cmd))

        # allow for 3 times the default spacing
        if len(commands):
            limit = formatter.width - 6 - max(len(cmd[0]) for cmd in commands)

            rows = []
            for subcommand, cmd in commands:
                help = cmd.get_short_help_str(limit)
                rows.append((subcommand, help))

            if rows:
                with formatter.section("Commands"):
                    formatter.write_dl(rows)

    def resolve_command(self, ctx, args):
        # always return the full command name
        _, cmd, args = super().resolve_command(ctx, args)
        return cmd.name, cmd, args


# https://stackoverflow.com/questions/50499340/specify-options-and-arguments-dynamically
class OptionEatAll(click.Option):
    def __init__(self, *args, **kwargs):
        self.save_other_options = kwargs.pop("save_other_options", True)
        nargs = kwargs.pop("nargs", -1)
        assert nargs == -1, "nargs, if set, must be -1 not {}".format(nargs)
        super(OptionEatAll, self).__init__(*args, **kwargs)
        self._previous_parser_process = None
        self._eat_all_parser = None

    def add_to_parser(self, parser, ctx):
        def parser_process(value, state):
            # method to hook to the parser.process
            done = False
            value = [value]
            if self.save_other_options:
                # grab everything up to the next option
                while state.rargs and not done:
                    for prefix in self._eat_all_parser.prefixes:
                        if state.rargs[0].startswith(prefix):
                            done = True
                    if not done:
                        value.append(state.rargs.pop(0))
            else:
                # grab everything remaining
                value += state.rargs
                state.rargs[:] = []
            value = tuple(value)

            # call the actual process
            self._previous_parser_process(value, state)

        retval = super(OptionEatAll, self).add_to_parser(parser, ctx)
        for name in self.opts:
            our_parser = parser._long_opt.get(name) or parser._short_opt.get(name)
            if our_parser:
                self._eat_all_parser = our_parser
                self._previous_parser_process = our_parser.process
                our_parser.process = parser_process
                break
        return retval


def multi_options(options):
    map_to_types = dict(
        array=str,
        number=float,
        string=str,
    )

    def decorator(f):
        for opt_params in reversed(options):
            param_decls = (
                "--" + opt_params["long"],
                opt_params["name"],
            )
            if "short" in opt_params and not opt_params["short"] is None:
                param_decls = ("-" + opt_params["short"], *param_decls)

            attrs = dict(
                required=opt_params["required"],
                type=map_to_types.get(opt_params["type"], opt_params["type"]),
                help=opt_params.get("help", ""),
            )
            if opt_params["type"] == "array":
                attrs["cls"] = OptionEatAll
                attrs["nargs"] = -1
                attrs["multiple"] = True

            click.option(*param_decls, **attrs)(f)
        return f

    return decorator


def installoptions_to_cli_options() -> List[Dict[str, Union[bool, str, Any, None]]]:
    from gefyra.types import GefyraInstallOptions

    result = []
    for _field in fields(GefyraInstallOptions):
        _data = dict(
            name=_field.name,
            long=_field.name.replace("_", "-"),
            short=_field.metadata.get("short"),
            required=False,
            help=_field.metadata.get("help"),
            type=_field.metadata.get("type") or "string",
        )
        result.append(_data)
    return result


def parse_ip_port_map(ctx, param, ports: Tuple[str]) -> dict:
    def v(p: str):
        if not p.isnumeric():
            raise RuntimeError(f"Invalid port {p}. Please use integer numbers as port.")
        return p

    # port - port
    res = {}
    for value in ports:
        _value = value.split(":")
        if len(_value) == 2:
            res[v(_value[1])] = v(_value[0])
        elif len(_value) == 3:
            res[v(_value[2])] = (_value[0], v(_value[1]))
        else:
            raise ValueError("Invalid value for port mapping.")
    return res


def parse_env(ctx, param, envs: Tuple[str]) -> List[str]:
    res = []
    for env in envs:
        if "=" not in env:
            raise ValueError("Invalid value for env variable. Please use 'ENV=value'.")
        res.append(env)
    return res


def parse_workload(ctx, param, workload: str) -> str:
    MSG = (
        "Invalid workload format. Please provide the workload "
        "in the format 'type/name' or 'type/name/container-name'."
    )
    if not workload:
        return workload
    if "/" not in workload:
        raise ValueError(MSG)
    return workload


def check_connection_name(ctx, param, selected: Optional[str] = None) -> str:
    from gefyra import api

    if not selected:
        selected = "default"

    conn_list = api.list_connections()
    if not conn_list:
        raise click.UsageError(
            message="No Gefyra connection found. Please connect to a cluster first or run by "
            "providing a connection file or run 'gefyra up'."
        )
    conn_names = [conn.name for conn in conn_list]
    if selected and selected in conn_names:
        return selected
    elif selected:
        raise click.BadParameter(
            message=f"The connection name '{selected}' does not exist."
        )
    else:
        if "default" in conn_names and len(conn_names) == 1:
            connection_name = "default"
        else:
            raise click.MissingParameter(
                message="Please provide a connection name from: {conn_names}",
                param="connection-name",  # type: ignore
            )
        return connection_name



import logging
import socket
import time
from typing import Any, Dict, Iterable, TYPE_CHECKING

if TYPE_CHECKING:
    from gefyra.types import GefyraBridge

logger = logging.getLogger(__name__)


def is_port_free(port):
    """
    Check if a port is free on the current system.
    """
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        try:
            s.bind(("127.0.0.1", port))
            return True
        except OSError:
            return False


def get_workload_type(workload_type_str: str):
    POD = ["pod", "po", "pods"]
    DEPLOYMENT = ["deploy", "deployment", "deployments"]
    STATEFULSET = ["statefulset", "sts", "statefulsets"]
    VALID_TYPES = POD + DEPLOYMENT + STATEFULSET

    if workload_type_str not in VALID_TYPES:
        raise RuntimeError(
            f"Unknown workload type {workload_type_str}\nValid workload types include:"
            f" {', '.join(str(valid_type) for valid_type in VALID_TYPES)}"
        )

    if workload_type_str in POD:
        return "pod"
    elif workload_type_str in DEPLOYMENT:
        return "deployment"
    elif workload_type_str in STATEFULSET:
        return "statefulset"


def generate_env_dict_from_strings(env_vars: Iterable[str]) -> dict:
    return {k[0]: k[1] for k in [arg.split("=", 1) for arg in env_vars] if len(k) > 1}


def wrap_bridge(bridge: Dict[Any, Any]) -> "GefyraBridge":
    from gefyra.types import GefyraBridge

    return GefyraBridge(
        provider=bridge["provider"],
        name=bridge["metadata"]["name"],
        client_id=bridge["client"],
        local_container_ip=bridge["destinationIP"],
        port_mappings=bridge["portMappings"] or [],
        target_container=bridge["targetContainer"],
        target_namespace=bridge["targetNamespace"],
        target_pod=bridge["targetPod"],
        state=bridge["state"],
    )


def stopwatch(func):
    def wrapper(*args, **kwargs):
        tic = time.perf_counter()
        result = func(*args, **kwargs)
        toc = time.perf_counter()
        logger.debug(
            f"Operation time for '{func.__name__}(...)' was {(toc - tic) * 1000:0.4f}ms"
        )
        return result

    return wrapper



from . import namespace  # noqa
from . import rbac  # noqa
from . import webhook  # noqa
from . import service  # noqa
from . import deployment  # noqa

COMPONENTS = [namespace, rbac, webhook, deployment, service]



import logging
import os
import sys
from threading import Thread, Event
from typing import Dict, List, Optional, TYPE_CHECKING
from gefyra.cluster.utils import retrieve_pod_and_container

if TYPE_CHECKING:
    from docker.models.containers import Container

from gefyra.configuration import (
    ClientConfiguration,
)
from .utils import generate_env_dict_from_strings, stopwatch


logger = logging.getLogger(__name__)

stop_thread = Event()


def print_logs(container: "Container"):
    for logline in container.logs(stream=True):
        print(logline.decode("utf-8"), end="")


def check_input():
    line = sys.stdin.readline()
    if not line:
        stop_thread.set()


@stopwatch
def run(
    image: str,
    connection_name: str,
    name: str = "",
    command: str = "",
    volumes: Optional[List] = None,
    ports: Optional[Dict] = None,
    detach: bool = True,
    auto_remove: bool = False,
    namespace: str = "",
    env: Optional[List] = None,
    env_from: str = "",
) -> bool:
    from kubernetes.client import ApiException
    from docker.errors import APIError
    from gefyra.cluster.utils import get_env_from_pod_container
    from gefyra.local.bridge import deploy_app_container
    from gefyra.local.utils import (
        get_processed_paths,
    )
    from gefyra.local.cargo import probe_wireguard_connection

    config = ClientConfiguration(connection_name=connection_name)

    # #125: Fallback to namespace in kube config
    if not namespace:
        from kubernetes.config import kube_config

        _, active_context = kube_config.list_kube_config_contexts()
        namespace = active_context["context"].get("namespace") or "default"
        ns_source = "kubeconfig"
    else:
        ns_source = "--namespace argument"

    dns_search = (
        f"{namespace}.svc.cluster.local svc.cluster.local cluster.local k8s".split(" ")
    )
    #
    # Confirm the wireguard connection working
    #
    try:
        probe_wireguard_connection(config)
    except Exception as e:
        logger.error(e)
        return False

    if volumes:
        volumes = get_processed_paths(os.getcwd(), volumes)
    #
    # 1. get the ENV together a) from a K8s container b) from override
    #
    env_dict = {}
    try:
        if env_from:
            env_from_pod, env_from_container = retrieve_pod_and_container(
                env_from, namespace=namespace, config=config
            )
            logger.debug(f"Using ENV from {env_from_pod}/{env_from_container}")
            raw_env = get_env_from_pod_container(
                config, env_from_pod, namespace, env_from_container
            )
            logger.debug("ENV from pod/container is:\n" + raw_env)
            raw_env_vars = raw_env.split("\n")
            env_dict = generate_env_dict_from_strings(raw_env_vars)
    except ApiException as e:
        logger.error(f"Cannot copy environment from Pod: {e.reason} ({e.status}).")
        return False
    if env:
        env_overrides = generate_env_dict_from_strings(env)
        env_dict.update(env_overrides)

    #
    # 2. deploy the requested container to Gefyra
    #
    try:
        container = deploy_app_container(
            config=config,
            image=image,
            name=name,
            command=command,
            ports=ports,
            env=env_dict,
            dns_search=dns_search,
            auto_remove=auto_remove,
            volumes=volumes,
        )
    except APIError as e:
        if e.status_code == 409:
            logger.error(e.explanation)
            return True
        else:
            raise RuntimeError(e.explanation)

    logger.info(
        f"Container image '{', '.join(container.image.tags)}' started with name"
        f" '{container.name}' in Kubernetes namespace '{namespace}' (from {ns_source})"
    )
    if detach:
        return True
    else:
        try:
            logger.debug("Now printing out logs")
            t = Thread(target=print_logs, args=[container], daemon=True)
            t.start()
            input_thread = Thread(target=check_input, daemon=True)
            input_thread.start()
            while t.is_alive():
                if stop_thread.is_set():
                    logger.info(f"Detached from container: {name}")
                    return True
        except KeyboardInterrupt:
            container.stop()
            logger.info(f"Container stopped: {name}")
    return True



import click
from gefyra.cli.utils import AliasedGroup, standard_error_handler


@click.group("operator", cls=AliasedGroup, help="Manage operator installation")
@click.pass_context
def operator(ctx):
    pass


@operator.command(
    "update",
    help=("Update operator to latest or specific version."),
)
@click.option(
    "--version",
    help="Set specific version to update operator to.",
    required=False,
    type=str,
)
@standard_error_handler
def update(version):
    from gefyra import api

    api.operator.update(
        version=version,
    )



import logging
from time import sleep
from typing import List, Dict, TYPE_CHECKING

from gefyra.exceptions import CommandTimeoutError, GefyraBridgeError

if TYPE_CHECKING:
    from gefyra.configuration import ClientConfiguration
    from gefyra.types import GefyraBridge


from .utils import stopwatch, wrap_bridge

logger = logging.getLogger(__name__)


def get_pods_to_intercept(
    workload_name: str, workload_type: str, namespace: str, config
) -> Dict[str, List[str]]:
    from gefyra.cluster.resources import (
        get_pods_and_containers_for_pod_name,
        get_pods_and_containers_for_workload,
    )

    pods_to_intercept = {}
    if workload_type != "pod":
        pods_to_intercept.update(
            get_pods_and_containers_for_workload(
                config, workload_name, namespace, workload_type
            )
        )
    else:
        pods_to_intercept.update(
            get_pods_and_containers_for_pod_name(config, workload_name, namespace)
        )
    return pods_to_intercept


def check_workloads(
    pods_to_intercept,
    workload_type: str,
    workload_name: str,
    container_name: str,
    namespace: str,
    config,
):
    from gefyra.cluster.resources import check_pod_valid_for_bridge

    pod_names = pods_to_intercept.keys()
    if len(pod_names) == 0:
        raise Exception("Could not find any pod to bridge.")

    cleaned_names = ["-".join(name.split("-")[:-2]) for name in pod_names]

    if workload_type != "pod" and workload_name not in cleaned_names:
        raise RuntimeError(
            f"Could not find {workload_type}/{workload_name} to bridge. Available"
            f" {workload_type}: {', '.join(cleaned_names)}"
        )
    if container_name not in [
        container for c_list in pods_to_intercept.values() for container in c_list
    ]:
        raise RuntimeError(f"Could not find container {container_name} to bridge.")

    for name in pod_names:
        check_pod_valid_for_bridge(config, name, namespace, container_name)


@stopwatch
def bridge(
    name: str,
    ports: dict,
    target: str,
    namespace: str = "default",
    handle_probes: bool = True,
    timeout: int = 0,
    wait: bool = False,
    connection_name: str = "",
) -> List["GefyraBridge"]:
    from docker.errors import NotFound
    from gefyra.local.bridge import get_all_gefyrabridges
    from gefyra.configuration import ClientConfiguration

    config = ClientConfiguration(connection_name=connection_name)

    try:
        container = config.DOCKER.containers.get(name)
    except NotFound:
        raise GefyraBridgeError(f"Could not find target container '{name}'")

    port_mappings = [f"{key}:{value}" for key, value in ports.items()]

    try:
        local_container_ip = container.attrs["NetworkSettings"]["Networks"][
            config.NETWORK_NAME
        ]["IPAddress"]
    except KeyError:
        raise GefyraBridgeError(
            f"The target container '{name}' is not in Gefyra's network"
            f" {config.NETWORK_NAME}. Did you run 'gefyra up'?"
        ) from None

    try:
        _bits = list(filter(None, target.split("/")))
        workload_type, workload_name = _bits[0:2]
        container_name = _bits[2] if _bits[2:] else None
    except IndexError:
        raise GefyraBridgeError(
            "Invalid --target notation. Use"
            " <workload_type>/<workload_name>(/<container_name>)."
        ) from None

    pods_to_intercept = get_pods_to_intercept(
        workload_name=workload_name,
        workload_type=workload_type,
        namespace=namespace,
        config=config,
    )
    if not container_name:
        container_name = pods_to_intercept[list(pods_to_intercept.keys())[0]][0]

    ireq_base_name = f"{name}-to-{namespace}.{workload_type}.{workload_name}"

    check_workloads(
        pods_to_intercept,
        workload_type=workload_type,
        workload_name=workload_name,
        container_name=container_name,
        namespace=namespace,
        config=config,
    )

    if len(pods_to_intercept.keys()) > 1:
        use_index = True
    else:
        use_index = False

    from gefyra.local.bridge import (
        get_gbridge_body,
        handle_create_gefyrabridge,
    )

    ireqs = []
    for idx, pod in enumerate(pods_to_intercept):
        logger.info(f"Creating bridge for Pod {pod}")
        ireq_body = get_gbridge_body(
            config,
            name=f"{ireq_base_name}-{idx}" if use_index else ireq_base_name,
            destination_ip=local_container_ip,
            target_pod=pod,
            target_namespace=namespace,
            target_container=container_name,
            port_mappings=port_mappings,
            handle_probes=handle_probes,
        )
        ireq = handle_create_gefyrabridge(config, ireq_body, target)
        logger.debug(f"Bridge {ireq['metadata']['name']} created")
        ireqs.append(ireq)
    #
    # block until all bridges are in place
    #
    logger.info("Waiting for the bridge(s) to become active")

    bridges = {str(ireq["metadata"]["uid"]): False for ireq in ireqs}
    waiting_time = 0
    # timeout = 0  means no timeout
    if timeout:
        waiting_time = timeout
    while True and wait:
        # watch whether all relevant bridges have been established
        gefyra_bridges = get_all_gefyrabridges(config)
        for gefyra_bridge in gefyra_bridges:
            if (
                gefyra_bridge["metadata"]["uid"] in bridges.keys()
                and gefyra_bridge.get("state", "") == "ACTIVE"
            ):
                bridges[str(gefyra_bridge["metadata"]["uid"])] = True
                logger.info(
                    f"Bridge {gefyra_bridge['metadata']['name']} "
                    f"({sum(bridges.values())}/{len(ireqs)}) established."
                )
        if all(bridges.values()):
            break
        sleep(1)
        # Raise exception in case timeout is reached
        waiting_time -= 1
        if timeout and waiting_time <= 0:
            raise CommandTimeoutError("Timeout for bridging operation exceeded")
    if not wait:
        gefyra_bridges = get_all_gefyrabridges(config)
        return list(map(wrap_bridge, gefyra_bridges))
    else:
        logger.info("Following bridges have been established:")
        _bridges = list(map(wrap_bridge, gefyra_bridges))
        for gefyra_bridge in gefyra_bridges:
            for port in port_mappings:
                (
                    pod_name,
                    ns,
                ) = (
                    gefyra_bridge["targetPod"],
                    gefyra_bridge["targetNamespace"],
                )
                bridge_ports = port.split(":")
                container_port, pod_port = bridge_ports[0], bridge_ports[1]
                logger.info(
                    f"Bridge for pod {pod_name} in namespace {ns} on port {pod_port} "
                    f"to local container {container_name} on port {container_port}"
                )
        return _bridges


def wait_for_deletion(gefyra_bridges: List, config: "ClientConfiguration"):
    from kubernetes.watch import Watch

    w = Watch()
    deleted = []
    uids = [gefyra_bridge["metadata"]["uid"] for gefyra_bridge in gefyra_bridges]
    for event in w.stream(
        config.K8S_CUSTOM_OBJECT_API.list_namespaced_custom_object,
        namespace=config.NAMESPACE,
        group="gefyra.dev",
        version="v1",
        plural="gefyrabridges",
    ):
        if event["type"] == "DELETED":
            if event["object"]["metadata"]["uid"] in uids:
                deleted.append(event["object"]["metadata"]["uid"])
                if set(deleted) == set(uids):
                    break


@stopwatch
def unbridge(
    name: str,
    connection_name: str = "",
    wait: bool = False,
) -> bool:
    from gefyra.local.bridge import handle_delete_gefyrabridge
    from gefyra.configuration import ClientConfiguration

    config = ClientConfiguration(connection_name=connection_name)

    gefyra_bridge = handle_delete_gefyrabridge(config, name)
    if gefyra_bridge:
        if wait:
            wait_for_deletion([gefyra_bridge], config=config)
        logger.info(f"Bridge {name} removed")
    return True


@stopwatch
def unbridge_all(
    wait: bool = False,
    connection_name: str = "",
) -> bool:
    from gefyra.configuration import ClientConfiguration
    from gefyra.local.bridge import (
        handle_delete_gefyrabridge,
        get_all_gefyrabridges,
    )

    config = ClientConfiguration(connection_name=connection_name)

    gefyra_bridges = get_all_gefyrabridges(config)
    for gefyra_bridge in gefyra_bridges:
        name = gefyra_bridge["metadata"]["name"]
        logger.info(f"Removing Bridge {name}")
        handle_delete_gefyrabridge(config, name)
    if wait:
        wait_for_deletion(config=config, gefyra_bridges=gefyra_bridges)
    return True



import base64
import logging
import os
from pathlib import Path
import socket
import time

from typing import IO, List, Optional, TYPE_CHECKING
from gefyra.api.clients import get_client
from gefyra.exceptions import GefyraConnectionError
from gefyra.local.clients import handle_get_gefyraclient
from gefyra.local.minikube import detect_minikube_config
from .utils import stopwatch

if TYPE_CHECKING:
    from docker.models.networks import Network


from gefyra.configuration import ClientConfiguration, get_gefyra_config_location
from gefyra.local.cargo import (
    create_wireguard_config,
    get_cargo_ip_from_netaddress,
    probe_wireguard_connection,
)
from gefyra.local.networking import get_or_create_gefyra_network, handle_remove_network
from gefyra.local.utils import (
    compose_kubeconfig_for_serviceaccount,
    handle_docker_get_or_create_container,
)
from gefyra.types import (
    GefyraClient,
    GefyraClientConfig,
    GefyraClientState,
    GefyraConnectionItem,
)


logger = logging.getLogger(__name__)


@stopwatch
def connect(  # noqa: C901
    connection_name: str,
    client_config: Optional[IO],
    minikube_profile: Optional[str] = None,
    probe_timeout: int = 60,
) -> bool:
    import kubernetes
    import docker

    cargo_container = None
    # if this connection already exists, just restore it
    if connection_name in [conns.name for conns in list_connections()]:
        logger.debug(f"Restoring exinsting connection {connection_name}")
        config = ClientConfiguration(connection_name=connection_name)
        cargo_container = config.DOCKER.containers.get(config.CARGO_CONTAINER_NAME)
        client = get_client(config.CLIENT_ID, connection_name=config.CONNECTION_NAME)
    else:
        # connection does not exist, so create it
        if client_config is None:
            raise GefyraConnectionError(
                "Connection is not yet created and no client configuration has been provided."
            )
        logger.debug(f"Creating new connection {connection_name}")
        file_str = client_config.read()
        client_config.close()
        gclient_conf = GefyraClientConfig.from_json_str(file_str)

        # this kubeconfig is being used by the client to operate in the cluster
        kubeconfig_str = compose_kubeconfig_for_serviceaccount(
            gclient_conf.kubernetes_server,
            gclient_conf.ca_crt,
            "gefyra",
            base64.b64decode(gclient_conf.token).decode("utf-8"),
        )
        loc = os.path.join(
            get_gefyra_config_location(),
            f"{connection_name}.yaml",
        )
        with open(loc, "w") as f:
            f.write(kubeconfig_str)
            logger.info(f"Client kubeconfig saved to {loc}")

        if minikube_profile:
            logger.debug(f"Minikube profile detected: {minikube_profile}")
            mini_conf = detect_minikube_config(minikube_profile)
            logger.debug(mini_conf)
            gclient_conf.gefyra_server = (
                f"{mini_conf['cargo_endpoint_host']}:{mini_conf['cargo_endpoint_port']}"
            )

        config = ClientConfiguration(
            connection_name=connection_name,
            kube_config_file=Path(loc),
            client_id=gclient_conf.client_id,
            cargo_endpoint_host=gclient_conf.gefyra_server.split(":")[0],
            cargo_endpoint_port=gclient_conf.gefyra_server.split(":")[1],
            cargo_container_name=f"gefyra-cargo-{connection_name}",
        )

        gclient = handle_get_gefyraclient(config, gclient_conf.client_id)
        client = GefyraClient(gclient, config)

        config.CARGO_PROBE_TIMEOUT = probe_timeout

    _retry = 0
    while _retry < 10:
        gefyra_network = get_or_create_gefyra_network(config)
        try:
            client.activate_connection(
                gefyra_network.attrs["IPAM"]["Config"][0]["Subnet"]
            )
            break
        except kubernetes.client.exceptions.ApiException as e:
            _retry += 1
            if e.status == 500:
                logger.debug(f"Could not activate connection, retrying {_retry}/10...")
                # if the given subnet is taken in the cluster (by another client), recreate the network and try again
                # hopefully the IPAM config will give a new subnet
                gefyra_network.remove()
    else:
        raise GefyraConnectionError("Could not activate connection") from None

    # busy wait for the client to enter the ACTIVE state
    _i = 0
    while _i < config.CONNECTION_TIMEOUT:
        if client.state == GefyraClientState.ACTIVE:
            break
        else:
            _i += 1
            time.sleep(1)
    else:
        raise GefyraConnectionError("Could not activate connection") from None
    client.update()

    # since this connection was (re)activated, save the current wireguard config (again)
    wg_conf = os.path.join(
        get_gefyra_config_location(), f"{config.CONNECTION_NAME}.conf"
    )
    if not client.provider_config:
        raise GefyraConnectionError(
            "Could not get provider config for client"
        ) from None

    if config.CARGO_ENDPOINT is None:
        config.CARGO_ENDPOINT = client.provider_config.pendpoint
    logger.debug(config.CARGO_ENDPOINT)
    # busy wait to resolve the cargo endpoint, making sure it's actually resolvable from this host
    _i = 0
    while _i < config.CONNECTION_TIMEOUT:
        try:
            socket.gethostbyname_ex(config.CARGO_ENDPOINT.split(":")[0])
            break
        except (socket.gaierror, socket.herror):  # [Errno -2] Name or service not known
            logger.debug(
                f"Could not resolve host '{config.CARGO_ENDPOINT.split(':')[0]}', "
                f"retrying {_i}/{config.CONNECTION_TIMEOUT}..."
            )
            _i += 1
            time.sleep(1)
    else:
        raise GefyraConnectionError(
            f"Cannot resolve host '{config.CARGO_ENDPOINT.split(':')[0]}'."
        ) from None

    with open(wg_conf, "w") as f:
        f.write(
            create_wireguard_config(
                client.provider_config, config.CARGO_ENDPOINT, config.WIREGUARD_MTU
            )
        )

    cargo_ip_address = get_cargo_ip_from_netaddress(
        gefyra_network.attrs["IPAM"]["Config"][0]["Subnet"]
    )

    try:
        if not cargo_container:
            cargo_container = handle_docker_get_or_create_container(
                config,
                f"{config.CARGO_CONTAINER_NAME}",
                config.CARGO_IMAGE,
                detach=True,
                cap_add=["NET_ADMIN"],
                privileged=True,
                volumes=[
                    "/var/run/docker.sock:/var/run/docker.sock",
                    f"{wg_conf}:/config/wg0.conf",
                ],
                pid_mode="host",
            )

            if minikube_profile:
                mini_conf = detect_minikube_config(minikube_profile)
                if mini_conf["network_name"]:
                    logger.debug("Joining minikube network")
                    minikube_net: "Network" = config.DOCKER.networks.get(
                        mini_conf["network_name"]
                    )
                    minikube_net.connect(cargo_container)
            logger.debug(f"Cargo gefyra net ip address: {cargo_ip_address}")
            gefyra_network.connect(cargo_container, ipv4_address=cargo_ip_address)
        cargo_container.start()
        time.sleep(1)
    except docker.errors.APIError as e:
        try:
            cargo_container and cargo_container.remove()
        except docker.errors.APIError:
            pass
        raise GefyraConnectionError(f"Could not start Cargo container: {e}") from None

    # Confirm the wireguard connection working
    logger.debug("Checking wireguard connection")
    probe_wireguard_connection(config)
    return True


@stopwatch
def disconnect(connection_name: str) -> bool:
    import docker

    config = ClientConfiguration(connection_name=connection_name)
    client = get_client(config.CLIENT_ID, connection_name=connection_name)
    get_or_create_gefyra_network(config)
    try:
        cargo_container = config.DOCKER.containers.get(
            f"{config.CARGO_CONTAINER_NAME}",
        )
        cargo_container.stop()
    except docker.errors.NotFound:
        pass
    client.deactivate_connection()
    return True


@stopwatch
def list_connections() -> List[GefyraConnectionItem]:
    from gefyra.local import CARGO_LABEL, CONNECTION_NAME_LABEL, VERSION_LABEL

    config = ClientConfiguration()
    result = []
    containers = config.DOCKER.containers.list(
        all=True, filters={"label": f"{CARGO_LABEL[0]}={CARGO_LABEL[1]}"}
    )
    for cargo_container in containers:
        if cargo_container.status == "running":
            try:
                config = ClientConfiguration(cargo_container_name=cargo_container.name)
                config.CARGO_PROBE_TIMEOUT = 1  # don't wait too long for the probe
                probe_wireguard_connection(config)
                state = "running"
            except GefyraConnectionError:
                state = "error"
        else:
            state = "stopped"
        result.append(
            GefyraConnectionItem(
                **{
                    "name": cargo_container.labels.get(
                        CONNECTION_NAME_LABEL, "unknown"
                    ),
                    "version": cargo_container.labels.get(VERSION_LABEL, "unknown"),
                    "created": cargo_container.attrs.get("Created", "unknown"),
                    "status": state,
                }
            )
        )
    return result


@stopwatch
def remove_connection(connection_name: str):
    import docker

    config = ClientConfiguration(connection_name=connection_name)
    try:
        get_client(
            config.CLIENT_ID, connection_name=connection_name
        ).deactivate_connection()
    except Exception as e:  # noqa E722
        logger.debug(e)
        pass
    try:
        cargo_container = config.DOCKER.containers.get(
            f"{config.CARGO_CONTAINER_NAME}",
        )
        cargo_container.remove(force=True)
    except docker.errors.NotFound:
        pass
    handle_remove_network(config)
    try:
        # remove kubeconfig file
        os.remove(os.path.join(get_gefyra_config_location(), f"{connection_name}.yaml"))
    except OSError:
        pass
    try:
        # remove wireguard config file
        os.remove(
            os.path.join(get_gefyra_config_location(), f"{config.CONNECTION_NAME}.conf")
        )
    except OSError:
        pass



import logging
import os
from typing import Optional

import click
from click import pass_context

from gefyra import api
from gefyra.cli.connections import _manage_container_and_bridges
from gefyra.cli.utils import standard_error_handler
from gefyra.configuration import ClientConfiguration, get_gefyra_config_location
from gefyra.exceptions import ClientConfigurationError, GefyraConnectionError
from gefyra.types import StatusSummary

logger = logging.getLogger("gefyra")


def _check_and_install(
    config: ClientConfiguration,
    connection_name: str = "",
    preset: Optional[str] = None,
    bar=None,
) -> bool:
    status = api.status(connection_name=connection_name)

    if status.summary == StatusSummary.UP:
        bar()
        bar.title = "Gefyra is already installed"
        return True
    elif status.summary == StatusSummary.INCOMPLETE:
        logger.warning("Gefyra is not installed, but operating properly. Aborting.")
        return False
    else:  # status.summary == StatusSummary.DOWN:
        logger.debug(f"Preset {preset}")
        api.install(
            kubeconfig=config.KUBE_CONFIG_FILE,
            kubecontext=config.KUBE_CONTEXT,
            apply=True,
            wait=True,
            preset=preset,
        )
        return True


@click.command("up", help="Install Gefyra on a cluster and directly connect to it")
@click.option(
    "--minikube",
    help="Connect Gefyra to a Minikube cluster (accepts minikube profile name, default is 'minikube'))",
    type=str,
    is_flag=False,
    flag_value="minikube",  # if --minikube is used as flag, we default to profile 'minikube'
    required=False,
)
@click.option(
    "--preset",
    help=f"Set configs from a preset (available: {','.join(api.LB_PRESETS.keys())})",
    type=str,
)
@pass_context
@standard_error_handler
def cluster_up(ctx, minikube: Optional[str] = None, preset: Optional[str] = None):
    from alive_progress import alive_bar
    from gefyra.exceptions import GefyraClientAlreadyExists, ClientConfigurationError
    from time import sleep
    import os

    if minikube and preset:
        raise click.BadOptionUsage(
            option_name="preset",
            message="Cannot use minikube together with preset flag.",
        )

    client_id = "default"
    connection_name = "default"
    kubeconfig = ctx.obj["kubeconfig"]
    kubecontext = ctx.obj["context"]

    config = ClientConfiguration(kube_config_file=kubeconfig, kube_context=kubecontext)
    with alive_bar(
        4,
        title="Installing Gefyra to the cluster",
        bar="smooth",
        spinner="classic",
        stats=False,
        dual_line=True,
    ) as bar:
        # run a default install
        install_success = _check_and_install(
            config=config, connection_name=connection_name, preset=preset, bar=bar
        )
        if install_success:
            bar()
            bar.title = f"Creating a Gefyra client: {client_id}"
            # create a client
            try:
                client = api.add_clients(
                    client_id,
                    kubeconfig=config.KUBE_CONFIG_FILE,
                    kubecontext=config.KUBE_CONTEXT,
                )[0]
            except GefyraClientAlreadyExists:
                client = api.get_client(
                    client_id,
                    kubeconfig=config.KUBE_CONFIG_FILE,
                    kubecontext=config.KUBE_CONTEXT,
                )
        else:
            client = api.get_client(
                client_id,
                kubeconfig=config.KUBE_CONFIG_FILE,
                kubecontext=config.KUBE_CONTEXT,
            )
            bar()
        # write config with specific connection data (for minikube)
        if not preset or (preset and api.PRESET_TYPE_MAPPING.get(preset) == "local"):
            host = config.CARGO_ENDPOINT.split(":")[0]
        else:
            host = None
        bar()
        bar.title = "Waiting for the Gefyra client to enter 'waiting' state"
        # busy wait for the client to enter the waiting state
        _i = 0
        while _i < 20:
            try:
                json_str = api.write_client_file(
                    client_id=client.client_id,
                    kubeconfig=config.KUBE_CONFIG_FILE,
                    kubecontext=config.KUBE_CONTEXT,
                    host=host,
                )
                break
            except ClientConfigurationError as e:
                logger.warning(e)
                sleep(1)
                _i += 1
        else:
            raise ClientConfigurationError(
                f"Could not set up the client '{client_id}'. This is most probably a problem of Gefyra operator. \n"
                f"Try running 'gefyra up{' --preset ' + preset if preset else ''}' again after some time."
            )

        # create a temporary file with the client config
        loc = os.path.join(
            get_gefyra_config_location(),
            f"{connection_name}_client.json",
        )
        fh = open(loc, "w+")
        fh.write(json_str)
        fh.seek(0)
        bar()
        bar.title = f"Connecting local network '{config.NETWORK_NAME}' to the cluster (up to 10 min)"
        logger.debug(f"Minikube profile {minikube}")
        try:
            # setting the probe timeout to a much higher value
            api.connect(
                connection_name,
                client_config=fh,
                minikube_profile=minikube,
                probe_timeout=180,
            )
        except GefyraConnectionError as e:
            raise GefyraConnectionError(
                f"Gefyra could not successfully establish the connection to '{config.CARGO_ENDPOINT.split(':')[0]}'.\n"
                "If you have run 'gefyra up' with a remote cluster, a newly created route may not be working "
                "immediately.\n"
                f"Try running 'gefyra up{' --preset ' + preset if preset else ''}' again after some time. "
                f"Error: {e}"
            ) from None
        fh.close()
        os.remove(loc)
        bar()
        bar.title = "Gefyra is ready"


@click.command("down", help="Remove Gefyra locally and on the cluster")
@pass_context
@standard_error_handler
def cluster_down(ctx):
    from alive_progress import alive_bar
    from gefyra import api

    if ctx.obj["kubeconfig"] is None:
        ctx.obj["kubeconfig"] = os.environ.get("KUBECONFIG") or os.path.expanduser(
            "~/.kube/config"
        )
    connection_name = "default"
    kubeconfig = ctx.obj["kubeconfig"]
    kubecontext = ctx.obj["context"]

    config = ClientConfiguration(kube_config_file=kubeconfig, kube_context=kubecontext)

    with alive_bar(
        2,
        title="Removing Gefyra from the cluster",
        bar="smooth",
        spinner="classic",
        stats=False,
        dual_line=True,
    ) as bar:
        try:
            _manage_container_and_bridges(connection_name=connection_name, force=True)
        except ClientConfigurationError:
            pass
        api.uninstall(
            kubeconfig=config.KUBE_CONFIG_FILE,
            kubecontext=config.KUBE_CONTEXT,
        )
        bar()
        bar.title = "Removing Gefyra from the local machine"
        api.remove_connection(connection_name=connection_name)
        bar()
        bar.title = "Gefyra successfully removed"
